////////////////////////////////////////////////////////////////////////////////
//
// TMED 1.0
// Author: Lubomir Havran
// Email: lhavran@dantech.sk
// Website: http://www.dantech.sk/
// Describe and features:
// This component detects motion from two images.
// - motion detection (2 events OnMotion, OnNotMotion)
// - edge detection - shows detected image
// - shows detected cells - scans red, green, blue separately or scans in grayscale
// - allows to define sensitivity for each cell
// - FREEWARE!!!
//
////////////////////////////////////////////////////////////////////////////////
unit MED;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, Math;

type
  TScanTypes = (sctRed, sctGreen, sctBlue, sctGrayscale);
  TScanType = set of TScanTypes;

  TShowItems = (shiGrid,shiEdges,shiDetectedCell);

  TShowItem = set of TShowItems;

  TMatrixItem = record
    red: cardinal;
    green: cardinal;
    blue: cardinal;
    sum: integer;
  end;

  TScanMatrix = array of TMatrixItem;

  TMotionItem = record
    RGB: TRGBTriple;
    Intensity: integer;
    Falloff: integer;
  end;

  TMotionItems = array of TMotionItem;

  PMotionItems = ^TMotionItems;

  TWindowScale = (wscStretched, wscNormal);

  TMED = class(TCustomControl)
  private
    Motion1: TMotionItems;
    Motion2: TMotionItems;
    ScanMatrix: TScanMatrix;
    FBitmap, EBitmap: TBitmap;
    fColumns: integer;
    fRows: integer;
    fScanType: TScanType;
    fGlobalScan: boolean;
    fGlobalSensitivity: integer;
    fOnMotion: TNotifyEvent;
    fisfirst: boolean;
    fWindowScale: TWindowScale;
    fScaning: boolean;
    fShowItem: TShowItem;
    fFalloff: integer;
    fOnNotMotion: TNotifyEvent;
    fLastDetected: boolean;
    fEdgeSensitivity: integer;
    fAbout: string;
    procedure SetFalloff(const Value: integer);
    procedure SetColumns(const Value: integer);
    procedure SetRows(const Value: integer);
    procedure SetAbout(const Value: string);
  private
    property isfirst: boolean read fisfirst write fisfirst;
    { Private declarations }
  protected
    procedure Paint; override;
    { Protected declarations }
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetMotionIntensity(MRow, MColumn: integer; MIntensity: integer);
    function GetMotionIntensity(MRow, MColumn: integer): integer;
    function InsertNextBitmap(NBitmap: TBitmap): integer;
    property Scaning: boolean read fScaning write fScaning;
    property LastDetected: boolean read fLastDetected;
    { Public declarations }
  published
    property About: string read fAbout write SetAbout;
    property Columns: integer read fColumns write SetColumns;
    property Rows: integer read fRows write SetRows;
    property Falloff: integer read fFalloff write SetFalloff;
    property ScanType: TScanType read fScanType write fScanType;
    property ShowItem: TShowItem read fShowItem write fShowItem;
    property GlobalSensitivity: integer read fGlobalSensitivity write fGlobalSensitivity;
    property EdgeSensitivity: integer read fEdgeSensitivity write fEdgeSensitivity;
    property GlobalScan: boolean read fGlobalScan write fGlobalScan;
    property OnMotion: TNotifyEvent read fOnMotion write fOnMotion;
    property OnNotMotion: TNotifyEvent read fOnNotMotion write fOnNotMotion;
    property WindowScale: TWindowScale read fWindowScale write fWindowScale;
    { Published declarations }
  published
    property Align;
    property Color;
    property Visible;
    property OnMouseMove;
    property OnMouseUp;
    property OnMouseDown;
    property OnClick;
    property OnDblClick;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Dantech', [TMED]);
end;

{ TMED }

constructor TMED.Create(AOwner: TComponent);
var
  x,y: integer;
begin
  inherited;
  fAbout:='(c) 2005 Dantech';
  Parent:=AOwner as TWinControl;
  FBitmap:=TBitmap.Create;
  FBitmap.PixelFormat:=pf24bit;
  FBitmap.Canvas.Pen.Color:=clGray;
  EBitmap:=TBitmap.Create;
  EBitmap.PixelFormat:=pf24bit;
  Rows:=25;
  Columns:=25;
  for x:=1 to Rows do begin
    for y:=1 to Columns do begin
      Motion1[(x-1)+(y-1)*Columns].Intensity:=5;
      Motion2[(x-1)+(y-1)*Columns].Intensity:=5;
    end;
  end;
  EdgeSensitivity:=10;
  GlobalSensitivity:=5;
  GlobalScan:=True;
  ScanType:=[sctRed]+[sctGreen]+[sctBlue];
  ShowItem:=[shiGrid]+[shiEdges]+[shiDetectedCell];
  Falloff:=0;
  width:=320;
  height:=240;
  Falloff:=1;
end;

destructor TMED.Destroy;
begin
  inherited;

end;

function TMED.GetMotionIntensity(MRow, MColumn: integer): integer;
begin
  result:=-1;
  if (MRow>=0) and (MRow<Rows) and (MColumn>=0) and (MColumn<Columns) then begin
    result:=Motion1[MRow+MColumn*Columns].Intensity;
  end;
end;

procedure TMED.SetMotionIntensity(MRow, MColumn, MIntensity: integer);
begin
  if (MRow>=0) and (MRow<Rows) and (MColumn>=0) and (MColumn<Columns) and
     (MIntensity>=0) and (MIntensity<=9) then begin
    Motion1[MRow+MColumn*Columns].Intensity:=MIntensity;
    Motion2[MRow+MColumn*Columns].Intensity:=MIntensity;
  end;

end;

function TMED.InsertNextBitmap(NBitmap: TBitmap): integer;
var
  ps,pd: PByteArray;
  sx,sy,ax,ay,fx,fy,x,y: integer;
  fw,fh: integer;
  fr,fg,fb: integer;
  frgb,frgb2: TRGBTriple;
  src,dst: PMotionItems;
  roz,r1,r2,r3: integer;
  wasm: boolean;
begin
  FBitmap.Width:=Width;
  FBitmap.Height:=Height;
  FBitmap.Canvas.Brush.Color:=clBlack;
  if not assigned(NBitmap) then exit;
  if (columns<1) or (rows<1) then exit;

  wasm:=False;
  Scaning:=True;
  SetLength(Motion1,rows*columns);
  SetLength(Motion2,rows*columns);
  SetLength(ScanMatrix,rows*columns);
  if WindowScale=wscStretched then
    FBitmap.Canvas.StretchDraw(RECT(0,0,Width,Height),NBitmap);
  if WindowScale=wscNormal then begin
    FBitmap.Canvas.FillRect(Rect(0,0,width,Height));
    FBitmap.Canvas.Draw(0,0,NBitmap);
  end;


  if (shiEdges in ShowItem) then begin
    EBitmap.Assign(FBitmap);
  end;

  //if ShowItem=[shiGrid] then begin
    for y:=0 to Rows-1 do begin
      for x:=0 to Columns-1 do begin
        fx:=x;
        fy:=y;
        ScanMatrix[fx+fy*Columns].red:=0;
        ScanMatrix[fx+fy*Columns].green:=0;
        ScanMatrix[fx+fy*Columns].blue:=0;
        ScanMatrix[fx+fy*Columns].sum:=0;
      end;
    end;
    pd:=nil;
    for y:=0 to FBitmap.height-1 do begin
      ps:=FBitmap.Scanline[y];
      if (shiEdges in ShowItem) then pd:=EBitmap.ScanLine[y];
      for x:=0 to FBitmap.width-1 do begin
        if (shiEdges in ShowItem) then begin
        if x=0 then begin
          pd[x*3]:=255;
          pd[x*3+1]:=255;
          pd[x*3+2]:=255;
        end else begin
          r1:=(ps[x*3]-ps[(x-1)*3]);
          r2:=(ps[x*3+1]-ps[(x-1)*3+1]);
          r3:=(ps[x*3+2]-ps[(x-1)*3+2]);
          roz:=round(sqrt(sqr(r1)+sqr(r2)+sqr(r3)));
          if roz>EdgeSensitivity then begin
            pd[x*3]:=0;
            pd[x*3+1]:=0;
            pd[x*3+2]:=0;
          end else begin
            pd[x*3]:=255;
            pd[x*3+1]:=255;
            pd[x*3+2]:=255;
          end;
        end;
        end;
        fx:=trunc((Columns/Width)*x);
        fy:=trunc((Rows/Height)*y);
        ScanMatrix[fx+fy*Columns].red:=ScanMatrix[fx+fy*Columns].red+ps[x*3+2];
        ScanMatrix[fx+fy*Columns].green:=ScanMatrix[fx+fy*Columns].green+ps[x*3+1];
        ScanMatrix[fx+fy*Columns].blue:=ScanMatrix[fx+fy*Columns].blue+ps[x*3];
        ScanMatrix[fx+fy*Columns].sum:=ScanMatrix[fx+fy*Columns].sum+1;
      end;
    end;

    if isfirst then begin
      src:=@Motion1;
      dst:=@Motion2;
    end else begin
      src:=@Motion2;
      dst:=@Motion1;
    end;

    if (shiEdges in ShowItem) then FBitmap.Assign(EBitmap);
    for y:=0 to Rows-1 do begin
      for x:=0 to Columns-1 do begin
        fx:=x;
        fy:=y;
        if ScanMatrix[fx+fy*Columns].sum>0 then begin
          dst^[fx+fy*Columns].RGB.rgbtRed:=round(ScanMatrix[fx+fy*Columns].red/ScanMatrix[fx+fy*Columns].sum);
          dst^[fx+fy*Columns].RGB.rgbtGreen:=round(ScanMatrix[fx+fy*Columns].green/ScanMatrix[fx+fy*Columns].sum);
          dst^[fx+fy*Columns].RGB.rgbtBlue:=round(ScanMatrix[fx+fy*Columns].blue/ScanMatrix[fx+fy*Columns].sum);
        end;
        ScanMatrix[fx+fy*Columns].red:=0;
        ScanMatrix[fx+fy*Columns].green:=0;
        ScanMatrix[fx+fy*Columns].blue:=0;
        ScanMatrix[fx+fy*Columns].sum:=0;

//        roz:=0;
        if (sctRed in ScanType) then
          r1:=(dst^[fx+fy*Columns].RGB.rgbtRed-src^[fx+fy*Columns].RGB.rgbtRed)
        else r1:=0;

        if (sctGreen in ScanType) then
          r2:=(dst^[fx+fy*Columns].RGB.rgbtGreen-src^[fx+fy*Columns].RGB.rgbtGreen)
        else r2:=0;

        if (sctBlue in ScanType) then
          r3:=(dst^[fx+fy*Columns].RGB.rgbtBlue-src^[fx+fy*Columns].RGB.rgbtBlue)
        else r3:=0;

        if (sctGrayscale in ScanType) then begin
          roz:=round(sqrt(sqr(r1)+sqr(r2)+sqr(r3)));
        end else begin
          r1:=INTEGER(77*src^[fx+fy*Columns].RGB.rgbtRed
          + 150*src^[fx+fy*Columns].RGB.rgbtGreen
          + 29*src^[fx+fy*Columns].RGB.rgbtBlue) shr 8;
          r2:=INTEGER(77*dst^[fx+fy*Columns].RGB.rgbtRed
          + 150*dst^[fx+fy*Columns].RGB.rgbtGreen
          + 29*dst^[fx+fy*Columns].RGB.rgbtBlue) shr 8;
          roz:=round(abs(r1-r2)*1.725);
        end;

        fw:=trunc(Width/Columns);
        fh:=trunc(Height/Rows);

        src^[fx+fy*Columns].Falloff:=src^[fx+fy*Columns].Falloff-1;
        dst^[fx+fy*Columns].Falloff:=dst^[fx+fy*Columns].Falloff-1;

        if (GlobalScan) and (roz>=(450-log10(GlobalSensitivity+1)*430)) then begin
          wasm:=True;
          src^[fx+fy*Columns].Falloff:=Falloff;
          dst^[fx+fy*Columns].Falloff:=Falloff;
        end;

        if (not GlobalScan) and (roz>=(450-log10(src^[fx+fy*Columns].Intensity+1)*430)) then begin
          wasm:=True;
          src^[fx+fy*Columns].Falloff:=Falloff;
          dst^[fx+fy*Columns].Falloff:=Falloff;
        end;

        if src^[fx+fy*Columns].Falloff<0 then begin
          src^[fx+fy*Columns].Falloff:=0;
          dst^[fx+fy*Columns].Falloff:=0;
        end;

        if (src^[fx+fy*Columns].Falloff>0) and (shiDetectedCell in ShowItem) then begin
          for ay:=trunc((Height/Rows)*fy) to trunc((Height/Rows)*fy)+fh do begin
            sy:=ay;
            if sy>(height-1) then sy:=height-1;
            pd:=FBitmap.Scanline[sy];
            for ax:=trunc((Width/Columns)*fx) to trunc((Width/Columns)*fx)+fw do begin
              sx:=ax;
              if sx>(width-1) then sx:=width-1;
              if (shiEdges in ShowItem) then begin
                fr:=255;
                fg:=255;
                fb:=pd[sx*3];;
              end else begin
                fr:=round(pd[sx*3+2]*1.5);
                fg:=round(pd[sx*3+1]*1.5);
                fb:=pd[sx*3];
              end;
              if fr>255 then fr:=255;
              if fg>255 then fg:=255;
              if fr>fg then fg:=fr
              else fr:=fg;
              pd[sx*3+2]:=fr;
              pd[sx*3+1]:=fg;
              pd[sx*3]:=fb;
            end;
          end;
        end;
      end;
    end;

  FBitmap.Canvas.Pen.Color:=clGray;
  if (shiGrid in ShowItem) then begin
    for x:=1 to Rows do begin
      FBitmap.Canvas.MoveTo(0,trunc((Height/Rows)*x));
      FBitmap.Canvas.LineTo(width-1,trunc((Height/Rows)*x));
    end;
    for y:=1 to Columns do begin
      FBitmap.Canvas.MoveTo(trunc((Width/Columns)*y),0);
      FBitmap.Canvas.LineTo(trunc((Width/Columns)*y),Height-1);
    end;
  end;

  fLastDetected:=wasm;
  isfirst:=not isfirst;
  Paint;
  if assigned(OnMotion) and wasm then OnMotion(self);
  if assigned(OnNotMotion) and (not wasm) then OnNotMotion(self);
  Scaning:=False;
end;

procedure TMED.Paint;
begin
  inherited;
  Canvas.Draw(0,0,FBitmap);
end;

procedure TMED.SetFalloff(const Value: integer);
begin
  if Value>0 then fFalloff := Value;
end;

procedure TMED.SetColumns(const Value: integer);
begin
  fColumns := Value;
  SetLength(Motion1,Rows*Columns);
  SetLength(Motion2,Rows*Columns);
end;

procedure TMED.SetRows(const Value: integer);
begin
  fRows := Value;
  SetLength(Motion1,Rows*Columns);
  SetLength(Motion2,Rows*Columns);
end;

procedure TMED.SetAbout(const Value: string);
begin
  fAbout := fAbout;
end;

end.
