////////////////////////////////////////////////////////////////////////////////
//
// TMED 1.0
// Author: Lubomir Havran
// Email: lhavran@dantech.sk
// Website: http://www.dantech.sk/
// Describe and features:
// This component detects motion from two images.
// - motion detection (2 events OnMotion, OnNotMotion)
// - edge detection - shows detected image
// - shows detected cells - scans red, green, blue separately or scans in grayscale
// - allows to define sensitivity for each cell
// - FREEWARE!!!
//
////////////////////////////////////////////////////////////////////////////////
Unit MED;

Interface

Uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, Math;

Type
  TScanTypes = (sctRed, sctGreen, sctBlue, sctGrayscale);
  TScanType  = Set Of TScanTypes;

  TShowItems = (shiGrid, shiEdges, shiDetectedCell);

  TShowItem = Set Of TShowItems;

  TMatrixItem = Record
    red:   Cardinal;
    green: Cardinal;
    blue:  Cardinal;
    sum:   Integer;
  End;

  TScanMatrix = Array Of TMatrixItem;

  TMotionItem = Record
    RGB:       TRGBTriple;
    Intensity: Integer;
    Falloff:   Integer;
  End;

  TMotionItems = Array Of TMotionItem;

  PMotionItems = ^TMotionItems;

  TWindowScale = (wscStretched, wscNormal);

  TMED = Class(TCustomControl)
  Private
    Motion1:            TMotionItems;
    Motion2:            TMotionItems;
    ScanMatrix:         TScanMatrix;
    FBitmap, EBitmap:   TBitmap;
    fColumns:           Integer;
    fRows:              Integer;
    fScanType:          TScanType;
    fGlobalScan:        Boolean;
    fGlobalSensitivity: Integer;
    fOnMotion:          TNotifyEvent;
    fisfirst:           Boolean;
    fWindowScale:       TWindowScale;
    fScaning:           Boolean;
    fShowItem:          TShowItem;
    fFalloff:           Integer;
    fOnNotMotion:       TNotifyEvent;
    fLastDetected:      Boolean;
    fEdgeSensitivity:   Integer;
    fAbout:             String;
    Procedure SetFalloff(Const Value: Integer);
    Procedure SetColumns(Const Value: Integer);
    Procedure SetRows(Const Value: Integer);
    Procedure SetAbout(Const Value: String);
  Private
    Property isfirst: Boolean Read fisfirst Write fisfirst;
    { Private declarations }
  Protected
    Procedure Paint; Override;
    { Protected declarations }
  Public
    Constructor Create(AOwner: TComponent); Override;
    Destructor Destroy; Override;
    Procedure SetMotionIntensity(MRow, MColumn: Integer; MIntensity: Integer);
    Function GetMotionIntensity(MRow, MColumn: Integer): Integer;
    Function InsertNextBitmap(NBitmap: TBitmap): Integer;
    Property Scaning: Boolean Read fScaning Write fScaning;
    Property LastDetected: Boolean Read fLastDetected;
    { Public declarations }
  Published
    Property About: String Read fAbout Write SetAbout;
    Property Columns: Integer Read fColumns Write SetColumns;
    Property Rows: Integer Read fRows Write SetRows;
    Property Falloff: Integer Read fFalloff Write SetFalloff;
    Property ScanType: TScanType Read fScanType Write fScanType;
    Property ShowItem: TShowItem Read fShowItem Write fShowItem;
    Property GlobalSensitivity: Integer Read fGlobalSensitivity Write fGlobalSensitivity;
    Property EdgeSensitivity: Integer Read fEdgeSensitivity Write fEdgeSensitivity;
    Property GlobalScan: Boolean Read fGlobalScan Write fGlobalScan;
    Property OnMotion: TNotifyEvent Read fOnMotion Write fOnMotion;
    Property OnNotMotion: TNotifyEvent Read fOnNotMotion Write fOnNotMotion;
    Property WindowScale: TWindowScale Read fWindowScale Write fWindowScale;
    { Published declarations }
  End;

Procedure Register;

Implementation

Procedure Register;
Begin
  RegisterComponents('Dantech', [TMED]);
End;

{ TMED }

Constructor TMED.Create(AOwner: TComponent);
Var
  x, y: Integer;
Begin
  Inherited;
  fAbout                   := '(c) 2005 Dantech';
  Parent                   := AOwner As TWinControl;
  FBitmap                  := TBitmap.Create;
  FBitmap.PixelFormat      := pf24bit;
  FBitmap.Canvas.Pen.Color := clGray;
  EBitmap                  := TBitmap.Create;
  EBitmap.PixelFormat      := pf24bit;
  Rows                     := 25;
  Columns                  := 25;
  For x := 1 To Rows Do
  Begin
    For y := 1 To Columns Do
    Begin
      Motion1[(x - 1) + (y - 1) * Columns].Intensity := 5;
      Motion2[(x - 1) + (y - 1) * Columns].Intensity := 5;
    End;
  End;
  EdgeSensitivity   := 10;
  GlobalSensitivity := 5;
  GlobalScan        := True;
  ScanType          := [sctRed] + [sctGreen] + [sctBlue];
  ShowItem          := [shiGrid] + [shiEdges] + [shiDetectedCell];
  Falloff           := 0;
  Width             := 320;
  Height            := 240;
  Falloff           := 1;
End;

Destructor TMED.Destroy;
Begin
  Inherited;

End;

Function TMED.GetMotionIntensity(MRow, MColumn: Integer): Integer;
Begin
  Result := -1;
  If (MRow >= 0) And (MRow < Rows) And (MColumn >= 0) And (MColumn < Columns) Then
  Begin
    Result := Motion1[MRow + MColumn * Columns].Intensity;
  End;
End;

Procedure TMED.SetMotionIntensity(MRow, MColumn, MIntensity: Integer);
Begin
  If (MRow >= 0) And (MRow < Rows) And (MColumn >= 0) And (MColumn < Columns) And
    (MIntensity >= 0) And (MIntensity <= 9) Then
  Begin
    Motion1[MRow + MColumn * Columns].Intensity := MIntensity;
    Motion2[MRow + MColumn * Columns].Intensity := MIntensity;
  End;

End;

Function TMED.InsertNextBitmap(NBitmap: TBitmap): Integer;
Var
  ps, pd:                       PByteArray;
  sx, sy, ax, ay, fx, fy, x, y: Integer;
  fw, fh:                       Integer;
  fr, fg, fb:                   Integer;
  frgb, frgb2:                  TRGBTriple;
  src, dst:                     PMotionItems;
  roz, r1, r2, r3:              Integer;
  wasm:                         Boolean;
Begin
  FBitmap.Width              := Width;
  FBitmap.Height             := Height;
  FBitmap.Canvas.Brush.Color := clBlack;
  If Not assigned(NBitmap) Then
    Exit;
  If (columns < 1) Or (rows < 1) Then
    Exit;

  wasm    := False;
  Scaning := True;
  SetLength(Motion1, rows * columns);
  SetLength(Motion2, rows * columns);
  SetLength(ScanMatrix, rows * columns);
  If WindowScale = wscStretched Then
    FBitmap.Canvas.StretchDraw(RECT(0, 0, Width, Height), NBitmap);
  If WindowScale = wscNormal Then
  Begin
    FBitmap.Canvas.FillRect(Rect(0, 0, Width, Height));
    FBitmap.Canvas.Draw(0, 0, NBitmap);
  End;


  If (shiEdges In ShowItem) Then
  Begin
    EBitmap.Assign(FBitmap);
  End;

  //if ShowItem=[shiGrid] then begin
  For y := 0 To Rows - 1 Do
  Begin
    For x := 0 To Columns - 1 Do
    Begin
      fx                                  := x;
      fy                                  := y;
      ScanMatrix[fx + fy * Columns].red   := 0;
      ScanMatrix[fx + fy * Columns].green := 0;
      ScanMatrix[fx + fy * Columns].blue  := 0;
      ScanMatrix[fx + fy * Columns].sum   := 0;
    End;
  End;
  pd := NIL;
  For y := 0 To FBitmap.Height - 1 Do
  Begin
    ps := FBitmap.Scanline[y];
    If (shiEdges In ShowItem) Then
      pd := EBitmap.ScanLine[y];
    For x := 0 To FBitmap.Width - 1 Do
    Begin
      If (shiEdges In ShowItem) Then
      Begin
        If x = 0 Then
        Begin
          pd[x * 3]     := 255;
          pd[x * 3 + 1] := 255;
          pd[x * 3 + 2] := 255;
        End
        Else
        Begin
          r1  := (ps[x * 3] - ps[(x - 1) * 3]);
          r2  := (ps[x * 3 + 1] - ps[(x - 1) * 3 + 1]);
          r3  := (ps[x * 3 + 2] - ps[(x - 1) * 3 + 2]);
          roz := round(sqrt(sqr(r1) + sqr(r2) + sqr(r3)));
          If roz > EdgeSensitivity Then
          Begin
            pd[x * 3]     := 0;
            pd[x * 3 + 1] := 0;
            pd[x * 3 + 2] := 0;
          End
          Else
          Begin
            pd[x * 3]     := 255;
            pd[x * 3 + 1] := 255;
            pd[x * 3 + 2] := 255;
          End;
        End;
      End;
      fx                                  := trunc((Columns / Width) * x);
      fy                                  := trunc((Rows / Height) * y);
      ScanMatrix[fx + fy * Columns].red   := ScanMatrix[fx + fy * Columns].red + ps[x * 3 + 2];
      ScanMatrix[fx + fy * Columns].green := ScanMatrix[fx + fy * Columns].green + ps[x * 3 + 1];
      ScanMatrix[fx + fy * Columns].blue  := ScanMatrix[fx + fy * Columns].blue + ps[x * 3];
      ScanMatrix[fx + fy * Columns].sum   := ScanMatrix[fx + fy * Columns].sum + 1;
    End;
  End;

  If isfirst Then
  Begin
    src := @Motion1;
    dst := @Motion2;
  End
  Else
  Begin
    src := @Motion2;
    dst := @Motion1;
  End;

  If (shiEdges In ShowItem) Then
    FBitmap.Assign(EBitmap);
  For y := 0 To Rows - 1 Do
  Begin
    For x := 0 To Columns - 1 Do
    Begin
      fx := x;
      fy := y;
      If ScanMatrix[fx + fy * Columns].sum > 0 Then
      Begin
        dst^[fx + fy * Columns].RGB.rgbtRed   := round(ScanMatrix[fx + fy * Columns].red / ScanMatrix[fx + fy * Columns].sum);
        dst^[fx + fy * Columns].RGB.rgbtGreen := round(ScanMatrix[fx + fy * Columns].green / ScanMatrix[fx + fy * Columns].sum);
        dst^[fx + fy * Columns].RGB.rgbtBlue  := round(ScanMatrix[fx + fy * Columns].blue / ScanMatrix[fx + fy * Columns].sum);
      End;
      ScanMatrix[fx + fy * Columns].red   := 0;
      ScanMatrix[fx + fy * Columns].green := 0;
      ScanMatrix[fx + fy * Columns].blue  := 0;
      ScanMatrix[fx + fy * Columns].sum   := 0;

      roz := 0;
      If (sctRed In ScanType) Then
        r1 := (dst^[fx + fy * Columns].RGB.rgbtRed - src^[fx + fy * Columns].RGB.rgbtRed)
      Else
        r1 := 0;

      If (sctGreen In ScanType) Then
        r2 := (dst^[fx + fy * Columns].RGB.rgbtGreen - src^[fx + fy * Columns].RGB.rgbtGreen)
      Else
        r2 := 0;

      If (sctBlue In ScanType) Then
        r3 := (dst^[fx + fy * Columns].RGB.rgbtBlue - src^[fx + fy * Columns].RGB.rgbtBlue)
      Else
        r3 := 0;

      If (sctGrayscale In ScanType) Then
      Begin
        roz := round(sqrt(sqr(r1) + sqr(r2) + sqr(r3)));
      End
      Else
      Begin
        r1  := Integer(77 * src^[fx + fy * Columns].RGB.rgbtRed + 150 * src^[fx + fy * Columns].RGB.rgbtGreen + 29 * src^[fx + fy * Columns].RGB.rgbtBlue) Shr 8;
        r2  := Integer(77 * dst^[fx + fy * Columns].RGB.rgbtRed + 150 * dst^[fx + fy * Columns].RGB.rgbtGreen + 29 * dst^[fx + fy * Columns].RGB.rgbtBlue) Shr 8;
        roz := round(abs(r1 - r2) * 1.725);
      End;

      fw := trunc(Width / Columns);
      fh := trunc(Height / Rows);

      src^[fx + fy * Columns].Falloff := src^[fx + fy * Columns].Falloff - 1;
      dst^[fx + fy * Columns].Falloff := dst^[fx + fy * Columns].Falloff - 1;

      If (GlobalScan) And (roz >= (450 - log10(GlobalSensitivity + 1) * 430)) Then
      Begin
        wasm                            := True;
        src^[fx + fy * Columns].Falloff := Falloff;
        dst^[fx + fy * Columns].Falloff := Falloff;
      End;

      If (Not GlobalScan) And (roz >= (450 - log10(src^[fx + fy * Columns].Intensity + 1) * 430)) Then
      Begin
        wasm                            := True;
        src^[fx + fy * Columns].Falloff := Falloff;
        dst^[fx + fy * Columns].Falloff := Falloff;
      End;

      If src^[fx + fy * Columns].Falloff < 0 Then
      Begin
        src^[fx + fy * Columns].Falloff := 0;
        dst^[fx + fy * Columns].Falloff := 0;
      End;

      If (src^[fx + fy * Columns].Falloff > 0) And (shiDetectedCell In ShowItem) Then
      Begin
        For ay := trunc((Height / Rows) * fy) To trunc((Height / Rows) * fy) + fh Do
        Begin
          sy := ay;
          If sy > (Height - 1) Then
            sy := Height - 1;
          pd := FBitmap.Scanline[sy];
          For ax := trunc((Width / Columns) * fx) To trunc((Width / Columns) * fx) + fw Do
          Begin
            sx := ax;
            If sx > (Width - 1) Then
              sx := Width - 1;
            If (shiEdges In ShowItem) Then
            Begin
              fr := 255;
              fg := 255;
              fb := pd[sx * 3];
              ;
            End
            Else
            Begin
              fr := round(pd[sx * 3 + 2] * 1.5);
              fg := round(pd[sx * 3 + 1] * 1.5);
              fb := pd[sx * 3];
            End;
            If fr > 255 Then
              fr := 255;
            If fg > 255 Then
              fg := 255;
            If fr > fg Then
              fg := fr
            Else
              fr := fg;
            pd[sx * 3 + 2] := fr;
            pd[sx * 3 + 1] := fg;
            pd[sx * 3]     := fb;
          End;
        End;
      End;
    End;
  End;

  FBitmap.Canvas.Pen.Color := clGray;
  If (shiGrid In ShowItem) Then
  Begin
    For x := 1 To Rows Do
    Begin
      FBitmap.Canvas.MoveTo(0, trunc((Height / Rows) * x));
      FBitmap.Canvas.LineTo(Width - 1, trunc((Height / Rows) * x));
    End;
    For y := 1 To Columns Do
    Begin
      FBitmap.Canvas.MoveTo(trunc((Width / Columns) * y), 0);
      FBitmap.Canvas.LineTo(trunc((Width / Columns) * y), Height - 1);
    End;
  End;

  fLastDetected := wasm;
  isfirst       := Not isfirst;
  Paint;
  If assigned(OnMotion) And wasm Then
    OnMotion(self);
  If assigned(OnNotMotion) And (Not wasm) Then
    OnNotMotion(self);
  Scaning := False;
End;

Procedure TMED.Paint;
Begin
  Inherited;
  Canvas.Draw(0, 0, FBitmap);
End;

Procedure TMED.SetFalloff(Const Value: Integer);
Begin
  If Value > 0 Then
    fFalloff := Value;
End;

Procedure TMED.SetColumns(Const Value: Integer);
Begin
  fColumns := Value;
  SetLength(Motion1, Rows * Columns);
  SetLength(Motion2, Rows * Columns);
End;

Procedure TMED.SetRows(Const Value: Integer);
Begin
  fRows := Value;
  SetLength(Motion1, Rows * Columns);
  SetLength(Motion2, Rows * Columns);
End;

Procedure TMED.SetAbout(Const Value: String);
Begin
  fAbout := fAbout;
End;

End.

